import glob,os
import logging
from posixpath import expanduser
from tools import fix_coords
from utils import get_ldmonths
import xarray as xr
from datetime import datetime,timedelta
import pandas as pd
import numpy as np

def get_obs(conf,target_var):
    
    src = conf.arch.get_system(conf.src_sys, conf.temp_freq)
    target = conf.arch.get_system(conf.target_sys, conf.temp_freq)
    logging.info(f"Loading observational datasets:{conf.src_sys} & {conf.target_sys}")

    training_years = range(conf.training_years[0],conf.training_years[-1]+1) 

    flatten = lambda t: [item for sublist in t for item in sublist]
    files = flatten([glob.glob(f"{src['dir']}/{src['subdirs'][var]}/{var}_{year}*.nc") 
                    for var in conf.src_vars[target_var] 
                    for year in training_years]) 


    srcobs = xr.open_mfdataset(files,
                               preprocess = fix_coords,
                               drop_variables='time_bnds',
                               combine="by_coords",
                               join = 'left')

    #-----------------------------
    # Load target
    #-----------------------------

    files = flatten([glob.glob(f"{target['dir']}/{target['subdirs'][target_var]}/{target_var}_{year}*.nc")
             for year in training_years])

    targetobs = xr.open_mfdataset(files,
                                  drop_variables='time_bnds',
                                  preprocess = fix_coords,
                                  combine="by_coords",
                                  join = 'left')
  
    return(targetobs,srcobs)

def preprocess_hcst(system):
    def preprocess(df):
        df = fix_coords(df)
        df = df.sel(dict(ensemble=range(system["ensemble"]["hcst"])))
        return(df)
    
    return(preprocess)

def check_missing_files(files):
    if files is None:
        raise Exception("There are no hcst files for the specified dataset")
    else:
        first_existing_file =  next(file for file in files if os.path.exists(file))
        new_files = []
        missing_files = []
        for file in files:
            if os.path.exists(file): 
                new_files.append(file)
                missing_files.append(False)
            else: 
                new_files.append(first_existing_file)
                missing_files.append(True)
    
    return(new_files,missing_files)


def get_hcst(conf,region,fcst_sdate,target_var):
    
    # --------------------------
    #       Hcst files load
    # -------------------------

    sys = conf.arch.get_system(conf.exp_sys,conf.temp_freq)
    hcst_years = range(conf.hcst_years[0],conf.hcst_years[-1]+1)

    sd_year = fcst_sdate[:4]
    sd_month = fcst_sdate[4:6]
    sd_day = fcst_sdate[6:8]

    exp_files = [f"{sys['dir']}/{sys['subdirs'][var]}/{var}_{str(year)}{sd_month}{sd_day}.nc"
                 for year in hcst_years
                 for var in conf.src_vars[target_var]]
    
    # detect missing files in hcst and replace by existing file 
    # (xarray does not allow loading as NAs the missing files)
    if conf.exp_sys == "cfsv2":
        exp_files,missing_files = check_missing_files(exp_files)
        not_missing_sdates = [idx for idx in range(len(missing_files)) 
                                            if not missing_files[idx]]
    else:
        not_missing_sdates = np.nan
    

    preprocess = preprocess_hcst(sys)

    exp = xr.open_mfdataset(exp_files,
                            preprocess=preprocess,
                            drop_variables=['time','time_bnds'],
                            combine='nested',
                            concat_dim='sdate')

    # Set to None the missing initially missing sdates 
    #if conf.exp_sys == "cfsv2":
    #    exp = exp.where(exp.sdate.isin(exp.sdate[not_missing_sdates]))

    exp['sdate'] = [f"{year}-{sd_month}-{sd_day}" for year in hcst_years]
    exp['sdate'] = [datetime.strptime(str(date.values), "%Y-%m-%d") for date in exp['sdate']]

    # goes from startdate to first lead day
    if sys['type'] == "seasonal":
        i_day = sd_day
        i_month = sd_month
    elif sys['type'] == "subseasonal":
        i_date = datetime.strptime(fcst_sdate, "%Y%m%d") + timedelta(days=4)
        i_day = i_date.strftime("%d")
        i_month = i_date.strftime("%m")

    leadtimes = [pd.date_range('{}-{}-{}'.format(year,i_month,i_day), 
                               periods=len(exp.time), freq='D') 
                for year in hcst_years]

    # New multidimensional coord leadtimes(sdate,time) with dates of each time step
    exp = exp.assign_coords(leadtimes=(('sdate', 'time'), leadtimes))


    # takes only the leadtimes needed
    # for subseasonal files are already cut to 4 weeks
    if sys['type'] == "seasonal":

        ld_months = get_ldmonths(conf.ldtimes[0],fcst_sdate[4:6])

        exp['time'] = pd.date_range('{}-{}-{}'.format('1900',sd_month,sd_day),
                      periods=len(exp.time), freq='D')
        exp = exp.where(exp.leadtimes.dt.month.isin(ld_months), drop=True)
        exp = exp.dropna(dim='time',how='any') # to drop 29th feb, look for another solution
        

    elif sys['type'] == "subseasonal":
        exp['time'] = pd.date_range('{}-{}-{}'.format(sd_year,i_month,i_day),
                      periods=len(conf.ldtimes[0])*7, freq='D')
    
    if not "time_bnds" in exp:
        exp['time_bnds'] = np.nan

    exp = exp.assign_attrs({'system':conf.exp_sys,"target_var":target_var,
                            'tempfreq':conf.temp_freq,'region':region,
                            'system_type':'hcst',"not_missing_sdates":not_missing_sdates})    
    return(exp)

def get_fcst(conf,region,fcst_sdate,target_var):


    sys = conf.arch.get_system(conf.exp_sys,conf.temp_freq)
    sd_year = fcst_sdate[:4]
    sd_month = fcst_sdate[4:6]
    sd_day = fcst_sdate[6:8]

    # --------------------------
    #       fcst files load
    # -------------------------

    ## TODO: work with mutiple src_vars
    exp_files = f"{sys['dir']}/{sys['subdirs'][target_var]}/{target_var}_{fcst_sdate}.nc"

    exp = xr.open_dataset(exp_files,
                          drop_variables=['time','bnds','realization'])

    exp = fix_coords(exp)

    # Creates sdate coord
    exp = exp.assign_coords(sdate=datetime.strptime(fcst_sdate,"%Y%m%d"))
    exp = exp.expand_dims('sdate')

    # goes from startdate to first lead day
    if sys['type'] == "seasonal":
        i_day = sd_day
        i_month = sd_month
        i_year = sd_year
    elif sys['type'] == "subseasonal":
        i_date = datetime.strptime(fcst_sdate, "%Y%m%d") + timedelta(days=4)
        i_day = i_date.strftime("%d")
        i_month = i_date.strftime("%m")
        i_year = i_date.strftime("%Y")

    # New multidimensional coord leadtimes(sdate,time) with dates of each time step
    leadtimes = pd.date_range('{}-{}-{}'.format(i_year,i_month,i_day), periods=len(exp.time), freq='D')
    exp = exp.assign_coords(leadtimes=(('time'), leadtimes))
    exp['time'] = leadtimes

    if sys['type'] == "seasonal":
        ld_months = get_ldmonths(conf.ldtimes[0],fcst_sdate[4:6])
        # filters timesteps corresponding to lead season
        exp = exp.where(exp.leadtimes.dt.month.isin(ld_months), drop=True)
        exp = exp.dropna(dim='time',how='any') # to drop 29th feb, look for another solution

    if not "time_bnds" in exp:
        exp['time_bnds'] = np.nan
    
    exp = exp.assign_attrs({'system':conf.exp_sys,"target_var":target_var,
                            'tempfreq':conf.temp_freq,'region':region,
                            'system_type':'fcst'})   
    
    return(exp)



def get_outfile(df,outdir):

    outdir = '{}/output/{}/{}/{}/'.format(outdir,df.attrs['system'],
                                          df.attrs['tempfreq'],
                                          df.attrs['target_var'])
    
    if df.attrs['system_type'] == "hcst":
    
        i_year = df.sdate[0].dt.strftime("%Y").values
        f_year = df.sdate[-1].dt.strftime("%Y").values
        mon_day = df.sdate[0].dt.strftime("%m%d").values
        outfile = f"{df.attrs['target_var']}-{df.attrs['region']}_{i_year}-{f_year}_{mon_day}.nc"

    elif df.attrs['system_type'] == "fcst":
        
        outfile = '{}-{}_{}.nc'.format(df.attrs['target_var'],
                                       df.attrs['region'],
                                       int(df.sdate.dt.strftime("%Y%m%d")))
    else:
        print("system_type not found")
                                                        
    return(outdir,outfile)

def save_nc(df,outdir):
    
    outdir,outfile = get_outfile(df,outdir)
    os.system(f'mkdir -p {outdir}')
    
    try:
        df = df.transpose("time","ensemble","bnds","sdate","lat","lon")
    except:
        df = df.transpose("time","ensemble","sdate","lat","lon")

    df.to_netcdf(outdir+outfile)

def _shif_times(df):

    times = []
    
    for leadtime in df.time.values:

        timestamp = pd.to_datetime(leadtime.astype(datetime)).to_pydatetime()
        if df.tempfreq in ["weekly", "weekly_mean"]:
            timestamp = timestamp #- timedelta(days=6)
        elif df.tempfreq in ["monthly", "monthly_mean"]:
            timestamp = timestamp.replace(day=1) 
        
        times.append(timestamp)

    return((['time'] ,times))

   
def get_time_bounds(df):

    bounds = []
    
    for leadtime in df.time.values:

        timestamp = pd.to_datetime(leadtime.astype(datetime)).to_pydatetime()
        if df.tempfreq in ["weekly", "weekly_mean"]:
            max_bound = timestamp
            min_bound = timestamp - timedelta(days=6)
        elif df.tempfreq in ["monthly", "monthly_mean"]:
            max_bound = timestamp
            min_bound = timestamp.replace(day=1) 
        
        bounds.append([min_bound, max_bound])

    return((['time', 'bnds'] ,bounds))

def get_time_bounds64(df):

    bounds = []
    
    for leadtime in df.time.values:

        if df.tempfreq in ["weekly", "weekly_mean"]:
            max_bound = leadtime
            min_bound = leadtime - np.timedelta64(6, 'D')
        elif df.tempfreq in ["monthly", "monthly_mean"]:
            max_bound = leadtime
            min_bound = np.datetime64(str(leadtime)[:8]+'01')
            

        
        bounds.append([min_bound, max_bound])

    return((['time', 'bnds'] ,bounds))

def _set_leadtimes(df):
    
    hcst_years = [str(year.values) for year in df.sdate.dt.year]
    
    # goes from startdate to first lead day
    if df.tempfreq in ["weekly", "weekly_mean"]:
        
        timestamp = pd.to_datetime(df.time[0].values.astype(datetime)).to_pydatetime()
        #i_date = timestamp + timedelta(days=4)
        i_day = timestamp.strftime("%d")
        i_month = timestamp.strftime("%m")
        leadtimes = [pd.date_range('{}-{}-{}'.format(year,i_month,i_day), 
                            periods=len(df.time), freq='7D') 
                    for year in hcst_years]
        
    elif df.tempfreq in ["monthly", "monthly_mean"]:
        i_day = str(df.time[0].dt.strftime("%d").values)
        i_month = str(df.time[0].dt.strftime("%m").values)
        leadtimes = [pd.date_range('{}-{}-{}'.format(year,i_month,i_day), 
                        periods=len(df.time), freq='1MS') 
                    for year in hcst_years]
    
    df = df.assign_coords(leadtimes=(('sdate', 'time'), leadtimes))

    return(df)

def to_monthly(df):
    
    logging.debug("Applying monthly mean to data")    
    df = df.resample(time='M', restore_coord_dims=True).mean(keep_attrs=True)
    df.attrs['tempfreq'] = 'monthly_mean'
    df['time_bnds'] = get_time_bounds64(df)
    df['time'] = _shif_times(df)
    df = _set_leadtimes(df)

    return(df)


def to_weekly(df):

    logging.debug("Applying weekly mean to data")    
    df = df.resample(time='7D', restore_coord_dims=True).mean(keep_attrs=True)
    df.attrs['tempfreq'] = 'weekly_mean'
    df['time_bnds'] = get_time_bounds64(df)
    df['time'] = _shif_times(df)
    df = _set_leadtimes(df)
 
    return(df)