#!/usr/bin/env python

import xarray as xr
import numpy as np
from sklearn.model_selection import train_test_split
import cv2
import sys

from utils import yml2range,load_fxconf,cut_domain,get_domain,fix_coords

def standardize(train, test, val, input_mean_std=None):
    """
    Standardizing per variable using global mean and std from the train dataset.
    """
    if input_mean_std is None:
        train_mean = np.mean(train, axis=(0,1,2))
        train_std = np.std(train, axis=(0))
        train_std = np.mean(train_std, axis=(0,1))
    else:
        train_mean = input_mean_std[0]
        train_std = input_mean_std[1]
    
    train = (train - train_mean) / train_std
    test = (test - train_mean) / train_std
    val = (val - train_mean) / train_std

    train = np.expand_dims(train, -1)
    test = np.expand_dims(test, -1)
    val = np.expand_dims(val, -1)
    return train, test, val, train_mean, train_std


def get_training_data(df, run_conf, varname='tas', test_size=0.4, random_state=11, return_mean_std=False, input_mean_std=None):
    """ """
    
    fx = load_fxconf(run_conf)
    YEARS = list(yml2range(fx['training_years']))
    YEARS_HOLDOUT = fx['holdout_years']
    YEARS_TRAINING = [i for i in YEARS if i not in YEARS_HOLDOUT]
    
    #_ = df.load()
    # Train split
    
    training_domain = get_domain("training",run_conf)
    if training_domain is not None:
        df = cut_domain(df,training_domain).load()

    df_holdout = df.sel(time=df.time.dt.year.isin(YEARS_HOLDOUT))
    df_training = df.sel(time=df.time.dt.year.isin(YEARS_TRAINING))
    
    
    x_train = df_training[varname].values
    x_test, x_val = train_test_split(df_holdout[varname].values, test_size=test_size, shuffle=True, random_state=random_state)
    # Standardizing
    x_train, x_test, x_val, x_train_mean, x_train_std = standardize(x_train, x_test, x_val, input_mean_std)
    print(x_train.shape, x_test.shape, x_val.shape)
    if return_mean_std:
        return x_train, x_test, x_val, x_train_mean, x_train_std
    else:
        return x_train, x_test, x_val


def load_imref(path, date='1983-01-01'):
    """ """
    dataset = xr.open_dataset(path)
     # Reference image
    imref = dataset.sel(time=date).tas.values
    
    return imref
    
    
def load_imref_era5preds(paths, date='1983-01-01'):
    """ 
    paths : tuple of str (paths)
    """
    preds = []
    for var in paths:
        dataset = xr.open_dataset(var)
        varnames = list(dataset.data_vars.keys())
        for i in varnames:
            if dataset[i].ndim == 3:
                varname = i
        predref = dataset[varname].sel(time=date).values
        dataset_training = dataset.sel(time=dataset.time.dt.year.isin(YEARS_TRAINING))
        x_train_mean = np.mean(dataset_training[varname].values)
        x_train_std = np.std(dataset_training[varname].values)
        predref = (predref - x_train_mean) / x_train_std
        preds.append(predref)
    
    arrpreds = np.asarray([np.squeeze(var) for var in preds])
    arrpreds = np.moveaxis(arrpreds, 0, -1)
    
    return arrpreds
    
def load_topo(df,shape,run_conf,path='/esarchive/scratch/cgomez/nbs/downscaling/data/GLOBE_land_elevation_CROPPED_SPAIN.nc'):
    """ 
    path global: '/esarchive/scratch/cgomez/nbs/downscaling/data/GLOBE_land_elevation.nc'
    """
    topo = xr.open_dataarray(path)
    topo = fix_coords(topo)
    training_domain = get_domain("training",run_conf)
    if training_domain is None:
        topo = cut_domain(topo, [df.lon.min().values,df.lon.max().values,df.lat.min().values,df.lat.max().values])
    else:
        topo = cut_domain(topo, training_domain)

    #newsize_xy=(df.lon.size, df.lat.size)
    newsize_xy = shape

    laoc = ~np.isnan(topo.values)
    laoc = cv2.resize(laoc.astype(int), newsize_xy, interpolation=cv2.INTER_NEAREST)

    topo = cv2.resize(topo.fillna(0).values, newsize_xy, interpolation=cv2.INTER_CUBIC)
    topo /= topo.max()

    return topo, laoc
